from json.decoder import JSONDecodeError
from .. import exceptions, settings
import requests


class RequestsTransport(object):
    def __init__(self, client_id, client_secret, access_token):
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = access_token

    def _request(self, method, path, *args, **kwargs):
        proto = 'https' if settings.SSL else 'http'
        url = '%s://%s/%s' % (proto, settings.DOMAIN, path)

        params = kwargs.pop('params', {})
        params['client_id'] = self.client_id

        if self.access_token:
            params['access_token'] = self.access_token

        method = getattr(requests, method.lower())
        response = method(
            url,
            params=params,
            headers={
                'Authorization': 'Bearer %s' % self.client_secret,
                'Content-Type': 'application/vnd.api+json',
                'Accepts': 'application/vnd.api+json'
            },
            *args,
            **kwargs
        )

        try:
            json = response.json()
        except JSONDecodeError:
            raise exceptions.BadResponseError(
                'Server did not reply with JSON-API data.'
            )

        if response.status_code == 400:
            raise exceptions.BadRequestError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 401:
            raise exceptions.UnauthorizedError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 403:
            raise exceptions.ForbiddenError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 404:
            raise exceptions.NotFoundError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 405:
            raise exceptions.MethodNotAllowedError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 406:
            raise exceptions.NotAcceptableError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 409:
            raise exceptions.ConflictError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 415:
            raise exceptions.UnsupportedMediaTypeError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code == 422:
            raise exceptions.UnprocessableEntityError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        if response.status_code < 200 or response.status_code >= 400:
            raise exceptions.UnknownError(
                json['error'].get('detail') or json['error'].get('title'),
                json['error'].get('meta')
            )

        return json

    def get(self, path, **kwargs):
        return self._request('GET', path, params=kwargs)

    def post(self, path, data={}):
        return self._request('POST', path, json=data)

    def patch(self, path, data={}):
        return self._request('PATCH', path, json=data)



class ConfigurationError(Exception):
    pass


class ClientError(Exception):
    pass


class UnauthorizedError(ClientError):
    pass


class UnprocessableEntityError(ClientError):
    pass


class BadRequestError(ClientError):
    pass


class ForbiddenError(ClientError):
    pass


class NotFoundError(ClientError):
    pass


class MethodNotAllowedError(ClientError):
    pass


class NotAcceptableError(ClientError):
    pass


class ConflictError(ClientError):
    pass


class UnsupportedMediaTypeError(ClientError):
    pass


class ServerError(Exception):
    pass


class InternalServerError(ServerError):
    pass


class BadResponseError(ServerError):
    pass


class UnknownError(ServerError):
    pass



import os


try:
    from django.conf import settings
except ImportError:
    settings = object()


def s(key, default=None):
    value = getattr(
        settings,
        'PODIANT_%s' % key,
        os.getenv('PODIANT_%s' % key)
    )

    return default if value is None else value


DOMAIN = s('DOMAIN', 'api.podiant.co')
SSL = s('SSL') != 'false'
CLIENT_ID = s('CLIENT_ID')
CLIENT_SECRET = s('CLIENT_SECRET')


__all__ = (
    'DOMAIN',
    'SSL',
    'CLIENT_ID',
    'CLIENT_SECRET'
)



from .api import Client
from .exceptions import UnknownError
import json


class WebhookMetaCollection(object):
    def __init__(self, data):
        for key, value in data.items():
            setattr(self, key, value)


class Webhook(object):
    def __init__(self, stream):
        data = json.loads(stream.decode('utf-8'))

        self.meta = WebhookMetaCollection(
            data.get('meta', {})
        )

        try:
            client = Client(data['meta']['access_token'])
            assert data['type'] == 'events'
            self.id = data['id']

            obj = data['attributes'].pop('object')
            assert 'type' in obj
            assert 'id' in obj

            collection = getattr(client, obj['type'])
            self.object = collection.filter(id=obj['id'])

            for key, value in data['attributes'].items():
                setattr(self, key, value)
        except (AssertionError, KeyError):
            raise UnknownError(
                'Data does not conform to expected Webhook object.'
            )



from datetime import date
from dateutil.parser import parse as parse_date
from types import GeneratorType
from urllib.parse import urlparse


class DictProxy(object):
    def __init__(self, request, data):
        self.__request = request
        self.__data = data

    def __getattr__(self, attr):
        try:
            return self.__data[attr]
        except KeyError:
            raise AttributeError(
                '\'%s\' has no attribute \'%s\'' % (
                    type(self).__name__,
                    attr
                )
            )


class AttributeProxy(object):
    def __init__(self, request, name):
        self.__locked = True
        self.__request = request
        self.__name = name
        self.__populated = False
        self.__value = None

    def __getvalue__(self):
        if not self.__populated:
            self.__request.__populate__()

        return self.__value

    def __setvalue__(self, value):
        self.__value = value
        self.__populated = True

    def __abs__(self):
        return self.__getvalue__().__abs__()

    def __and__(self, other):
        return self.__getvalue__().__and__(other)

    def __bool__(self):
        return bool(self.__getvalue__())

    def __ceil__(self):
        return self.__getvalue__().__ceil__()

    def __cmp__(self, other):
        return self.__getvalue__().__cmp__(other)

    def __dir__(self):
        return self.__getvalue__().__dir__()

    def __doc__(self):
        return self.__getvalue__().__doc__()

    def __eq__(self, other):
        return self.__getvalue__() == other

    def __float__(self):
        return self.__getvalue__().__float__()

    def __floor__(self):
        return self.__getvalue__().__floor__()

    def __format__(self, f):
        return self.__getvalue__().__format__(f)

    def __ge__(self, other):
        return self.__getvalue__() >= other

    def __get__(self, key):
        return self.__getvalue__().__get__(key)

    def __gt__(self, other):
        return self.__getvalue__() > other

    def __hash__(self):
        return self.__getvalue__().__hash__()

    def __int__(self):
        return int(self.__getvalue__())

    def __iter__(self):
        return self.__getvalue__().__iter__()

    def __le__(self, other):
        return self.__getvalue__() <= other

    def __len__(self):
        return len(self.__getvalue__())

    def __lt__(self, other):
        return self.__getvalue__() < other

    def __ne__(self, other):
        return self.__getvalue__() != other

    def __or__(self, other):
        return self.__getvalue__() or other

    def __pos__(self):
        return self.__getvalue__().__pos__()

    def __repr__(self):
        return self.__getvalue__().__repr__()

    def __round__(self, *args):
        return self.__getvalue__().__round__(*args)

    def __str__(self):
        return str(self.__getvalue__())

    def __getattr__(self, attr):
        value = self.__getvalue__()
        return getattr(value, attr)

    def __call__(self, **kwargs):
        path = '%s%s/' % (self.__request.path, self.__name)
        return self.__request.bundle.client.post(path, kwargs)

    def to(self, totype):
        try:
            converter = {
                'str': str,
                'unicode': str,
                'int': int,
                'float': float,
                'bool': bool,
                'list': list,
                'tuple': tuple,
                'set': set,
                'dict': dict,
                'date': parse_date,
                'datetime': parse_date
            }[totype]
        except KeyError:
            raise TypeError('Unknown type: \'%s\'', totype)

        return converter(self.__getvalue__())


class RelationshipProxy(object):
    def __init__(self, request, data):
        urlparts = urlparse(data['links']['related'])

        self.__path = urlparts.path[1:]
        self.__request = request
        self.__populated = False
        self.__value = None

        if 'data' in data:
            if not self.__populated:
                self.__value = []

            single = False
            if isinstance(data['data'], dict):
                data['data'] = [data['data']]
                single = True

            for item in data['data']:
                for inclusion in request.bundle.included:
                    if item['type'] == inclusion['type']:
                        if item['id'] == inclusion['id']:
                            cls_name = '%sEntity' % (
                                item['type'].replace(
                                    '-', ' '
                                ).title().replace(
                                    ' ', ''
                                )
                            )

                            cls = type(cls_name, (ObjectProxy,), {})
                            self.__value.append(
                                cls(
                                    bundle=request.bundle,
                                    path=inclusion['links']['self'],
                                    kind=item['type'],
                                    id=item['id'],
                                    attributes=inclusion['attributes'],
                                    links=inclusion['links'],
                                    meta=inclusion.get('meta', {}),
                                    relationships=inclusion.get(
                                        'relationships',
                                        {}
                                    )
                                )
                            )

            if single and any(self.__value):
                self.__value = self.__value[0]
                self.__populated = True
            elif not single:
                self.__populated = True

    def __getvalue__(self):
        if not self.__populated:
            self.__value = self.__request.bundle.client.get(self.__path)

        return self.__value

    def __len__(self):
        self.__getvalue__()

        if isinstance(self.__value, GeneratorType):
            self.__value = list(self.__value)

        return len(self.__value)

    def __iter__(self):
        return self.__getvalue__().__iter__()

    def __getattr__(self, attr):
        return getattr(self.__getvalue__(), attr)


class ObjectProxy(object):
    def __init__(
        self,
        bundle, path, query={},
        kind=None, id=None, attributes=None,
        links=None, relationships=None, meta=None
    ):
        self.__bundle = bundle
        self.__path = path
        self.__query = query
        self.__id = id
        self.__kind = kind
        self.__cache = {}
        self.__attributes = attributes if attributes is not None else {}
        self.__rels = relationships if relationships is not None else {}
        self.__dirty = {}

        if links is not None:
            cls_name = '%sLinkCollection' % (
                kind.replace('-', ' ').title().replace(' ', '')
            )

            cls = type(
                cls_name,
                (DictProxy,),
                {}
            )

            self.__links = cls(self, links)

        if meta is not None:
            cls_name = '%sMetaCollection' % (
                kind.replace('-', ' ').title().replace(' ', '')
            )

            cls = type(
                cls_name,
                (DictProxy,),
                {}
            )

            self.__meta = cls(self, meta)

    def __repr__(self):
        if self.__id is not None:
            return '<%s %s>' % (
                type(self).__name__,
                self.__id
            )

        return '<%s>' % type(self).__name__

    def __str__(self):
        return self.__repr__()

    def dict(self):
        if self.__id is None or self.__kind is None:
            self.__populate__()

        if self.__id is None:
            raise AttributeError(
                '\'%s\' has no attribute \'dict\'' % (
                    type(self).__name__
                )
            )

        d = dict(**self.__attributes)
        d['id'] = self.__id

        return d

    def include(self, *items):
        include = self.__query.get('include', '').split(',')
        include.extend(items)
        self.__query['include'] = ','.join(
            sorted(
                set(
                    [i for i in include if i and i.strip()]
                )
            )
        )

        return self

    def order_by(self, *fields):
        sort = self.__query.get('sort', '').split(',')
        sort.extend(fields)
        self.__query['sort'] = ','.join(
            sorted(
                set(
                    [s for s in sort if s and s.strip()]
                )
            )
        )

        return self

    def all(self):
        return self.filter()

    def __iter__(self):
        return self.all().__iter__()

    def filter(self, id=None, **kwargs):
        if id:
            Model = type(self)
            path = '%s%s/' % (self.__path, id)
            return Model(self.__bundle, path)

        query = dict(**self.__query)
        for key, value in kwargs.items():
            query['filter[%s]' % key] = value

        return self.__bundle.client.get(self.__path, **query)

    def __getattr__(self, attr):
        if attr == 'links':
            if self.__id is None and self.__kind is None:
                self.__populate__()

            return self.__links

        if attr == 'meta':
            if self.__id is None and self.__kind is None:
                self.__populate__()

            return self.__meta

        if attr in 'bundle':
            return self.__bundle

        if attr in 'path':
            return self.__path

        if attr in 'id':
            return self.__id

        if attr not in self.__cache:
            jsonapi_version = attr.replace('_', '-')
            self.__cache[attr] = AttributeProxy(self, jsonapi_version)

            if jsonapi_version in self.__attributes:
                self.__cache[attr].__setvalue__(
                    self.__attributes[jsonapi_version]
                )
            elif jsonapi_version in self.__rels:
                self.__cache[attr].__setvalue__(
                    RelationshipProxy(
                        self,
                        self.__rels[jsonapi_version]
                    )
                )

        return self.__cache[attr]

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return

        if self.__id is None and self.__kind is None:
            self.__populate__()

        jsonapi_version = attr.replace('_', '-')

        if jsonapi_version not in self.__attributes:
            raise AttributeError(
                '\'%s\' object has no attribute \'%s\'' % (
                    type(self).__name,
                    attr
                )
            )

        v = None
        if isinstance(value, date):
            v = value.isoformat()
        else:
            v = value

        self.__cache[attr] = AttributeProxy(
            self, jsonapi_version
        )

        self.__cache[attr].__setvalue__(v)
        self.__dirty[jsonapi_version] = v

    def __populate__(self, data=None):
        if data is None:
            if self.__id is not None or self.__kind is not None:
                return

            data = self.__bundle.client.load(self.__path, **self.__query)

        self.__kind = data['type']
        self.__id = data['id']
        self.__attributes = data['attributes']
        self.__rels = data.get('relationships', {})
        self.__dirty = {}

        for jsonapi_version, value in data['attributes'].items():
            python_version = jsonapi_version.replace('-', '_')
            if python_version not in self.__cache:
                self.__cache[python_version] = AttributeProxy(
                    self, jsonapi_version
                )

            self.__cache[python_version].__setvalue__(value)

        for jsonapi_version, value in data.get('relationships', {}).items():
            python_version = jsonapi_version.replace('-', '_')
            if python_version not in self.__cache:
                self.__cache[python_version] = AttributeProxy(
                    self, jsonapi_version
                )

            self.__cache[python_version].__setvalue__(
                RelationshipProxy(self, value)
            )

        cls_name = '%sLinkCollection' % (
            data['type'].replace('-', ' ').title().replace(' ', '')
        )

        cls = type(
            cls_name,
            (DictProxy,),
            {}
        )

        self.__links = cls(self, data['links'])

        cls_name = '%sMetaCollection' % (
            data['type'].replace('-', ' ').title().replace(' ', '')
        )

        cls = type(
            cls_name,
            (DictProxy,),
            {}
        )

        self.__meta = cls(self, data.get('meta'))

    def save(self):
        if not any(self.__dirty.keys()):
            pass

        data = self.__bundle.client.patch(self.__path, self.__dirty)
        self.__populate__(data)



from urllib.parse import urlparse, parse_qs
from .. import exceptions, settings
from .proxy import ObjectProxy
from .transport import RequestsTransport


class RequestBundle(object):
    def __init__(self, client, included):
        self.client = client
        self.included = included


class Client(object):
    def __init__(self, access_token=None, client_id=None, client_secret=None):
        if client_id is None:
            client_id = settings.CLIENT_ID

        if client_secret is None:
            client_secret = settings.CLIENT_SECRET

        if not client_id:
            raise exceptions.ConfigurationError(
                'Missing PODIANT_CLIENT_ID setting.'
            )

        if not client_secret:
            raise exceptions.ConfigurationError(
                'Missing PODIANT_CLIENT_SECRET setting.'
            )

        self.__cache = {}
        self.__transport = RequestsTransport(
            client_id,
            client_secret,
            access_token
        )

    def load(self, path, **query):
        response = self.__transport.get(path, **query)
        data = response['data']
        return data

    def get(self, path, **params):
        response = self.__transport.get(path, **params)
        data = response['data']

        if isinstance(data, list):
            while True:
                for datum in data:
                    yield self.__proxy__(
                        datum['type'],
                        datum['links']['self'],
                        id=datum['id'],
                        attributes=datum['attributes'],
                        links=datum['links'],
                        meta=datum.get('meta', {}),
                        relationships=datum.get('relationships', {}),
                        included=response.get('included')
                    )

                links = response.get('links', {})
                if links.get('next'):
                    query = parse_qs(urlparse(links['next']).query)
                    params['page'] = query['page'][0]
                    response = self.__transport.get(path, **params)
                    data = response['data']
                else:
                    break
        else:
            yield self.__proxy__(
                data['type'],
                data['links']['self'],
                id=data['id'],
                attributes=data['attributes'],
                links=data['links'],
                meta=data.get('meta', {}),
                relationships=data.get('relationships', {}),
                included=response.get('included')
            )

    def post(self, path, data):
        response = self.__transport.post(path, data)
        data = response['data']

        if isinstance(data, list):
            return [
                self.__proxy__(
                    datum['type'],
                    datum['links']['self'],
                    id=datum['id'],
                    attributes=datum['attributes'],
                    links=datum['links'],
                    meta=datum.get('meta', {}),
                    relationships=datum.get('relationships', {}),
                    included=response.get('included')
                ) for datum in data
            ]

        return self.__proxy__(
            data['type'],
            data['links']['self'],
            id=data['id'],
            attributes=data['attributes'],
            links=data['links'],
            meta=data.get('meta', {}),
            relationships=data.get('relationships', {}),
            included=response.get('included')
        )

    def patch(self, path, data):
        response = self.__transport.patch(path, data)
        data = response['data']

        return {
            'type': data['type'],
            'id': data['id'],
            'attributes': data['attributes'],
            'links': data['links'],
            'meta': data.get('meta', {}),
            'relationships': data.get('relationships', {}),
            'included': response.get('included')
        }

    def __proxy__(self, kind, path, qs={}, **kwargs):
        cls_name = '%sEntity' % (
            kind.replace('-', ' ').title().replace(' ', '')
        )

        cls = type(cls_name, (ObjectProxy,), {})
        included = kwargs.pop('included', {})
        return cls(
            bundle=RequestBundle(self, included),
            path=path,
            query=qs,
            kind=kind,
            **kwargs
        )

    def __getattr__(self, attr):
        kind = attr.replace('_', '-')
        if kind not in self.__cache:
            self.__cache[kind] = self.__proxy__(kind, '%s/' % kind)

        return self.__cache[kind]
